#!/usr/bin/Rscript

# Performs a Fisher's exact test on the nucleotide frequencies in 
# reference sequences (frA .. rfT) versus vaccine sequences (vfA .. vfT)

ALPHA <- 0.01	# significance level

suppressMessages(library(dplyr))
suppressMessages(library(statmod))

stdin <- file('stdin')
freqs <- tbl_df(read.delim(stdin))

# Computes Fisher's test on the mA, MA freqs in one table row
fisher.on.table.row <- function(row){
	ref_MA <- as.integer(row['ref_MA'])
	ref_mA <- as.integer(row['ref_mA'])
	vac_MA <- as.integer(row['vac_MA'])
	vac_mA <- as.integer(row['vac_mA'])
	
	m <- matrix(c(ref_MA, ref_mA, vac_MA, vac_mA), ncol=2)
	#print(m)
	ft <- fisher.test(m)
	# now compute the power of the test
	n_ref    <- ref_MA + ref_mA
	n_vac    <- vac_MA + vac_mA
	prop_ref <- ref_mA / n_ref
	prop_vac <- vac_mA / n_vac
	# power is undefined if one of the proportions is not defined
	if (0 == n_ref || 0 == n_vac) {
		power = NaN
	} else {
		power    <- power.fisher.test(prop_ref, prop_vac, n_ref, n_vac,
																	alpha=ALPHA)
	}

	return(c(n_ref, prop_ref, n_vac, prop_vac,
					 ft$p.val, ft$conf.int[1], ft$estimate, ft$conf.int[2],
					 power))
}

fisher.test.results <- apply(freqs, 1, fisher.on.table.row)
m <- as.matrix(fisher.test.results, ncol=9) %>% t()
colnames(m) <- c('n_ref', 'prop_ref', 'n_vac', 'prop_vac', 'p.val', 'CI.lo', 'OR', 'CI.hi', 'power')

# This is meant to be read back by find_mutations.py

# NaNs originally arise from 0/0 (while x/0, where x != 0, yield +-Inf). But
# NOTE: write.table() treats NaN as NA, which I find rather strange, since I
# thought NA stands for a missing value, while NaN means "not a number". The
# problem is that Python can convert NaN into a numeric value (i.e., nan), while
# NA (quite sensibly, I might add) causes a ValueError (try float('NaN') and
# float('NA')).  So I have to use na='NaN' to tell R to treat NaNs as NaNs...
# which will work as long as I don't have missing values. Sheesh.

write.table(as.data.frame(m), '', quote=FALSE, sep='\t', na='NaN')
