#!/usr/bin/env python3

# Takes a file of Genbank entries containing (hopefully) RSV sequence data, and
# writes a CSV file containing (in that order) ID, subtype, category, gene,
# sequence length, percent N, and sequence. Subtype is the virus subtype (i.e.,
# A or B); category is either WILD or VACCINE.

# The subtype, category and gene are inferred from annotation, and are therefore
# not 100% accurate. The program also tries to detect any sequences that are not
# human RSV, and rejects them. If output stream 3 is open, the program will
# print the rejected entries to it (useful for debugging).

from Bio import SeqIO
from sys import stdin, stderr, argv, exit
from warnings import warn
from os import fdopen

import re

from gene_name_map import controlled_names

def infer_gene(cds):
	gene = None
	if 'gene' in cds.qualifiers:
		gene = cds.qualifiers['gene'][0]
	elif 'product' in cds.qualifiers:
		gene = cds.qualifiers['product'][0]
	if gene in controlled_names:
		gene = controlled_names[gene]
	else:
		gene = 'UNKNOWN'
	return gene
	
def extract_CDS_data(seq, cds):
	'''Returns relevant data for the CDS passed as 'cds' (a SeqFeature) in a
	tuple. The 'seq' argument is used to extract the sequence.'''
	cds_seq = cds.extract(seq)
	gene = infer_gene(cds)
	return {'gene': gene, 'CDS': cds_seq}

def extract_CDS(record):
	cds = [ft for ft in record.features if ft.type == 'CDS']
	if not cds:
		# Record has no CDS annotations: just return whole seq, considering the
		# gene unknown,
		cds_data = [{'gene': 'UNKNOWN', 'CDS': record.seq}]
	else:
		# Record has CDS annotations: infer gene and extract CDS
		cds_data = [extract_CDS_data(record.seq, cds) for cds in cds]
	return cds_data

rsv_re = re.compile(r'respiratory syncytial virus', re.I)
arsv_re = re.compile(r'(\w+) respiratory syncytial virus', re.I)

def is_hRSV(record):
	'''Returns True IFF source organism is human RSV'''
	organism = record.annotations['source']
	if None is rsv_re.search(organism):
		# that's not even RSV
		return False
	else:
		# RSV, but which host?
		match = arsv_re.search(organism)
		if None is match:
			# just 'respiratory syncytial virus' (host unknown, assume human)
			return True
		else:
			if str.lower(match.group(1)) != 'human':
				# nonhuman host: reject
				return False
			else:
				return True

def look_for_keywords(keywords, string):
	'''
	Takes a list of keywords and loks for each of them in 'string'.
	Returns False if none of the keywords is found in the string, 
	and True otherwise.
	look_for_keywords(['one', 'two'], "three') => False
	look_for_keywords(['one', 'two'], "no one') => True
	'''
	for kw in keywords:
		if string.find(kw) >= 0:
			return True
	return False

def infer_category(record):
	'''
	Infers the record's category, i.e. vaccine/attenuated or wild. Returns
	'VACCINE' or 'WILD' accordingly.
	'''
	ann = record.annotations
	source = ann['source'].lower()
	# look for 'vaccine', 'attenuat', and more if needed
	target_kw = ['vaccine', 'attenuat', 'patent']
	if look_for_keywords(target_kw, source):
		return 'VACCINE'
	definition = record.description.lower()
	if look_for_keywords(target_kw, definition):
		return 'VACCINE'
	titles = "\n".join([ref.title for ref in ann['references']]).lower()
	if look_for_keywords(target_kw, titles):
		return 'VACCINE'
	if 'comment' in ann:
		comment = ann['comment'].lower()
		if look_for_keywords(target_kw, comment):
			return 'VACCINE'

	# No indication of either vaccine or patent: assume wild.
	return 'WILD'

def percent_n(seq):
	return 100 * seq.lower().count('n') / len(seq)

# NOTE: in all the regexps below, I specify that the subtype/group/strain/etc.
# must be exactly one letter long, i.e. 'B' but not 'B2'. This is because it
# seems a relatively reliable way of telling subtype names from genotype names,
# while the words type, subtype, group, subgroup, strain, etc. are much less
# reliable (see also infer_subtype() below).

# A regex that captures expressions like "subtype B", "group: A", etc.
subtype_re = re.compile(r'(group|type)[: ="]+([ab])\b', re.I)
# A regex that captures expressions like "hrva", "hrv-b", etc
rsvab_re = re.compile(r'rsv[ -]?([ab])\b', re.I)
# and one that catches fully spelled out names
rsvfullab_re = re.compile(r'respiratory syncytial virus ([ab])\b', re.I)
# and one for strains
strain_re = re.compile(r'strain[=" ]+([ab])\b', re.I)

def search_for_subtype(string):
	'''
	Searches for matches of any of the subtype-catching regexps.
	Returns the subtype, or None if not found.
	'''
	match = subtype_re.search(string)
	if match is not None:
		return str.upper(match.group(2))

	match = rsvab_re.search(string)
	if match is not None:
		return str.upper(match.group(1))

	match = rsvfullab_re.search(string)
	if match is not None:
		return str.upper(match.group(1))

	# this one seems less reliable, so I try it last.
	match = strain_re.search(string)
	if match is not None:
		return str.upper(match.group(1))

	return None

def infer_subtype(record):
	'''Infers the subtype (A or B) from record annotations. Returns 'A' or 'B',
	or 'UNKNOWN' if neither.
    We want to match strings like "type", "subtype", "subgroup" etc.
	WARNING: there is inconsistency in the use of these terms: "genotype" 
	is strictly speaking something else (genotype codes have two or three letters
	followed by a number, e.g. "GA1" or "SAB3"; but confusingly some entries are
	annotated "genotype: group A", which I will take to mean SUBtype A (there
	being no such thing, as we have just seen, as "genotype" A). Likewise, some
	entries use "group" to refer to genotype, and others use "group" to refer to
	subtype. Heck, some people even use "strain" for the same thing. The rule of
	thumb I use here is that subtypes have a single letter,
	while genotypes have at least two.
	See
	https://en.wikipedia.org/wiki/Human_respiratory_syncytial_virus.

	Returns a tuple containing the inferred subtype and an indication of the
	field it was inferred from.

	I checked this function by direct cross-checking of the inferred subtype
	versus the annotations, on a few dozen entries. No error was found.
	'''

	ann = record.annotations
	# first, see if the Source field has the subtype
	source = ann['source']
	source_subtype = search_for_subtype(source)
	if source_subtype is not None:
		return (source_subtype, 'SOURCE')
	# then check the Organism
	organism = ann['organism']
	organism_subtype = search_for_subtype(organism)
	if organism_subtype is not None:
		return (organism_subtype, 'ORGANISM')
	# then the Description
	description_subtype = search_for_subtype(record.description)
	if description_subtype is not None:
		return (description_subtype, 'DESCRIPTION')

	#for k, v in record.annotations.items(): print("%s -> %s" % (k, v))
	for ft in record.features:
		if ft.type == 'source':
			# First, check for an explicit serotype
			if 'serotype' in ft.qualifiers:
				cand_subtype = str.upper(
						ft.qualifiers['serotype'][0])
				if cand_subtype == 'A' or cand_subtype == 'B':
					return (cand_subtype, 'FT-SEROTYPE')
			# No luck... do a regex search on the values
			quals = "\n".join([q for qs in list(ft.qualifiers.values()) for q in qs])
			#print(quals)
			quals_subtype = search_for_subtype(quals)
			if quals_subtype is not None:
				return (quals_subtype, 'FT-QUALS')


	# I look last in reference titles, because it's quite conceivable that a
	# paper title says something like "RSV-A and RSV-B" (case in point:
	# KJ939944), which leaves us none the wiser.
	for ref in record.annotations['references']:
		title_subtype = search_for_subtype(ref.title)
		if title_subtype is not None:
			return (title_subtype, 'TITLE')
	
	return ('UNKNOWN', 'DEFAULT')

infile = stdin # I really prefer true filters
#infile = open(argv[1], 'r') # for running the debugger
rejected = []

print("ID\tSubtype\tCategory\tGene\tLength\tPctN\tCDS")
for record in SeqIO.parse(infile, 'genbank'):
	if not is_hRSV(record):
		rejected.append(record)
		continue
	category = infer_category(record)
	# To debug the origin of the subtype, just use the whole tuple:
	#subtype = infer_subtype(record)
	subtype = infer_subtype(record)[0]
	if category != 'WILD':
		cdss = [{'gene': 'UNKNOWN', 'CDS': record.seq}]
	else:
		cdss = extract_CDS(record)
	lines = ["%s\t%s\t%s\t%s\t%d\t%d\t%s" % (
		record.name, subtype, category, cds['gene'], 
		len(cds['CDS']), percent_n(cds['CDS']), cds['CDS']) for cds in cdss]
	for l in lines: print(l)

# If fd 3 is open, write the rejected entries to it.  To open fd 3, just use one
# of the shell operators (e.g. 3>reject), or do it beforehand with e.g. 'exec 3>
# reject'

try:
	fo = fdopen(3, "w")
	SeqIO.write(rejected, fo, 'genbank')
except OSError:
	pass

