#!/usr/bin/env python3
# -*- coding: iso-8859-15 -*-

from Bio import AlignIO
from Bio.Seq import Seq
from Bio.Data.CodonTable import TranslationError
from Bio.SubsMat import MatrixInfo
from sys import stdin, stdout
from re import search
from sys import exit, argv
from math import log
from subprocess import Popen, PIPE
from sqlite3 import connect
from os import fdopen
from collections import defaultdict

nucleotides = ("A", "C", "G", "T")


# Usage: find_mutations.py <pval-cutoff> < <aln>
# where aln is a FastA alignment file

def allele_counts(col):
    """
    Returns a residue -> count mapping (absolute frequencies)"
    """
    counts = {}
    for r in [c.upper() for c in col]:
        if r == "-": continue
        if r in counts:
            counts[r] = counts[r] + 1
        else:
            counts[r] = 1
    return counts

def allele_frequencies(counts):
    """
    Returns a residue -> frequency mapping (relative frequencies)
    counts is a residue -> count dictionary (see allele_counts())
    """
    freq = {}
    s = float(sum(counts.values()))
    for residue in list(counts.keys()):
        freq[residue] = counts[residue] / s
    return freq

def major_minor_allele_frequencies(ref_freqs, vac_freqs):
    '''
    Returns a tuple of major allele (MA), MA freq in refs (ref_MA), minor
    allele freq in refs (ref_mA), and ditto in vaccines (vac_MA; vac_mA).
    These are _absolute_ frequencies (i.e., counts).
    '''
    tot_freqs = {}
    for nt in nucleotides:
        tot_freqs[nt] = ref_freqs.get(nt,0) + vac_freqs.get(nt,0)
    nt_by_freq = sorted(tot_freqs.items(), key=lambda t: t[1])
    maj_A = nt_by_freq[-1][0]
    ref_MA = ref_freqs.get(maj_A,0)
    vac_MA = vac_freqs.get(maj_A,0)
    ref_mA = sum([v for (k, v) in ref_freqs.items() if k != maj_A])
    vac_mA = sum([v for (k, v) in vac_freqs.items() if k != maj_A])
    tot_ref = sum([v for (k, v) in ref_freqs.items()]) 
    tot_vac = sum([v for (k, v) in vac_freqs.items()]) 
    assert(ref_MA + ref_mA == tot_ref)
    assert(vac_MA + vac_mA == tot_vac)
    return {'MA': maj_A, 'ref_MA': ref_MA, 'ref_mA': ref_mA, 'vac_MA': vac_MA,
            'vac_mA': vac_mA}

def get_category(record, conn):
    csr = conn.cursor()
    csr.execute('SELECT Category FROM rsv WHERE ID = ?', (record.id,))
    return csr.fetchone()[0]

def odds(ref_freq, vac_freq):
    def odd(r, v):
        if r == 0:
            if v == 0:
                return 'Undef'
            else:
                return 'Inf'
        else:
            return "%g" % (v / r)
    return [odd(ref_freq.get(nt,0), vac_freq.get(nt,0)) for nt in nucleotides]

def get_codon_pos(n):
    '''
    Given a position in a coding nucleotide sequence, returns the corresponding
    position within the codon.
    E.g. get_codon_pos(1) -> 1
         get_codon_pos(5) -> 2
         get_codon_pos(9) -> 3
    This also works for gapped sequences, provided the number of gaps upstream
    of n is a multiple of 3.
    '''
    return (n-1) % 3 + 1

def get_codons_at(seqs, pos):
    '''
    Takes a set of aligned DNA sequences and a nucleotide position, and returns
    the codon at the corresponding position in all the sequences, taking frame
    into account. Of course it is assumed that te alignment starts in codon
    position 1 (ie, frame 1) and has no frameshifts.
    '''
    codon_pos = get_codon_pos(pos) # pos of nt within codon, ie 1, 2, or 3
    codon_start_pos = pos - codon_pos # starting at 0
    return [seq[codon_start_pos:codon_start_pos+3] for seq in seqs]

def codon_freq_at(seqs, pos):
    '''
    Takes a set of aligned DNA sequences and a nucleotide position, and returns
    the codon frequency at that position, according to get_codons_at().
    Gap codons are given a frequency of -1, so that sorting by frequency never 
    gives a gap codon on top.
    '''
    d = defaultdict(int)
    for cod in get_codons_at(seqs, pos):
        cod_str = str(cod)
        if cod_str == '---':
            d[cod_str] = -1
        else:
            d[cod_str] += 1
    return d

def subst_score(aa1, aa2):
    try:
        return MatrixInfo.blosum100[(aa1, aa2)]
    except KeyError: # (x,y) not found -> try (y,x)
        try:
            return MatrixInfo.blosum100[(aa2, aa1)]
        except KeyError: # STOP codon?
            return 'Stop'
    else:
        return 'Undef'

PVAL_CUTOFF = float(argv[1])
conn = connect('int/all.db')
aln = AlignIO.read(stdin, "fasta")

ref_rcds = [rec for rec in aln if get_category(rec, conn) == 'WILD']

ref_seqs = [record.seq for record in ref_rcds]
ref_cols = list(zip(*ref_seqs))   # each column as a tuple
ref_cnts = [allele_counts(col) for col in ref_cols]
ref_freq = [allele_frequencies(cnt) for cnt in ref_cnts]

vac_rcds = [rec for rec in aln if get_category(rec, conn) == 'VACCINE']

vac_seqs = [record.seq for record in vac_rcds]
vac_cols = list(zip(*vac_seqs))   # each column as a tuple
vac_cnts = [allele_counts(col) for col in vac_cols]
vac_freq = [allele_frequencies(cnt) for cnt in vac_cnts]

debug  = open('fisher.csv', 'w')

fisher = Popen('./bin/Fisher_test_on_nt_freqs.R',
        shell=True, stdin=PIPE, stdout=PIPE)

fisher_test_lines = ["pos\tref_A\tref_C\tref_G\tref_T\tvac_A\tvac_C\tvac_G\tvac_T\tMA\tref_MA\tref_mA\tvac_MA\tvac_mA\n"]
# Hey, Python can do basic pattern matching à la Haskell!
for i, (r, v) in enumerate(zip(ref_cnts, vac_cnts)):
    ref_cnts_w0 = [r.get(nt, 0) for nt in nucleotides]
    vac_cnts_w0 = [v.get(nt, 0) for nt in nucleotides]
    allele_freqs = major_minor_allele_frequencies(r, v)
    counts = "\t".join(map(str, ref_cnts_w0 + vac_cnts_w0))
    alleles = "{}\t{}\t{}\t{}\t{}".format(
        allele_freqs['MA'], allele_freqs['ref_MA'], allele_freqs['ref_mA'],
        allele_freqs['vac_MA'], allele_freqs['vac_mA'])
    line = "%d\t%s\t%s\n" % (i+1, counts, alleles) # +1: seqs start @1, Py @0
    fisher_test_lines.append(line)

for line in fisher_test_lines:
    fisher.stdin.write(str.encode(line))
    debug.write(line)

fisher_out = bytes.decode(fisher.communicate()[0])

with open('R_out', 'w') as r:
    r.write(fisher_out)

fisher_lines = fisher_out.split('\n')
print("Pos\tCPos\tModalRefCodon\tModalVacCodon\tModalRefAA\tModalVacAA\tModalSubstScore\tTopVacAA\tTopSubstScore\tN_ref\tmA_prop_ref\tN_vac\tmA_prop_vac\tP-val\tCI_low\tOR\tCI_high\tpower\tref_A\tref_C\tref_G\tref_T\tvac_A\tvac_C\tvac_G\tvac_T\tA_odds\tC_odds\tG_odds\tT_odds")
for line in fisher_lines[1:]:
    if len(line) == 0: continue
    lnum, n_ref, prop_ref, n_vac, prop_vac, ft_pval, ci_lo, odds_ratio, ci_hi, power = line.split()
    lnum, n_ref, prop_ref, n_vac, prop_vac, ft_pval, ci_lo, odds_ratio, ci_hi = \
        int(lnum), int(n_ref), float(prop_ref), int(n_vac), float(prop_vac), float(ft_pval), float(ci_lo), float(odds_ratio), float(ci_hi)
    if 'NA' != power:
        power = float(power)
    if ft_pval < PVAL_CUTOFF:
        counts = "\t".join(fisher_test_lines[lnum].split()[1:9])
        mut_odds = "\t".join(odds(ref_freq[lnum-1], vac_freq[lnum-1]))
        codon_pos = get_codon_pos(lnum)
        ref_codons_d = codon_freq_at(ref_seqs, lnum) 
        top_ref_codon = [cod for cod, freq in ref_codons_d.items() \
                if freq == max(ref_codons_d.values())][0]
        vac_codons_d = codon_freq_at(vac_seqs, lnum)
        top_vac_codon = [cod for cod, freq in vac_codons_d.items() \
                if freq == max(vac_codons_d.values())][0]
        try:
            top_ref_aa = str(Seq(top_ref_codon).translate())
        except TranslationError:
            top_ref_aa = '-'

        try:
            top_vac_aa = str(Seq(top_vac_codon).translate())
        except TranslationError:
            top_vac_aa = '-'

        if (top_ref_aa != '-' and top_vac_aa != '-'):
            score = subst_score(top_ref_aa, top_vac_aa)
        
        # get a set of all aa at the position in vaccines (not just the most 3
        # abundant one)
        vac_aa = set([ str(Seq(cod).translate()) for cod
                   in vac_codons_d if cod.find('-') < 0])
        # compute substitution scores from the most abundant aa in wild
        # sequences to each of the vaccines aa
        vac_aa_subst_scores = [ (aa, subst_score(top_ref_aa, aa))
                for aa in vac_aa ]
        vac_aa_subst_scores_num = [ tpl for tpl in vac_aa_subst_scores
                if type(tpl[1]) is int]
        # get that mutation which has the lowest score (i.e., most likely to
        # alter the phenotype)
        try:
            strongest_mut = (sorted(vac_aa_subst_scores_num, key=lambda mut: mut[1]))[0]
        except IndexError:
            strongest_mut = ('-', '-')
        #print("{} -> {} => {}".format(top_ref_aa, vac_aa_subst_scores, lowest_sc))

        print(lnum, codon_pos, top_ref_codon, top_vac_codon,
                top_ref_aa, top_vac_aa, score, strongest_mut[0],
                strongest_mut[1], n_ref, prop_ref, n_vac, prop_vac,
                ft_pval, ci_lo, odds_ratio, ci_hi, power,
                counts, mut_odds, 
                sep='\t')
